use crate::camera::Camera;
use crate::hittable::{Hittable, Sphere};
use crate::output::{Output, P3};
use crate::ray::Ray;
use crate::vec3::{Color, Point3, Vec3};
use rand::distributions::{Distribution, Uniform};
use crate::material::{Dielectric, Lambertian, Material, Metal};

mod vec3;
mod ray;
mod hittable;
mod material;
mod camera;
mod output;

fn random_scene() -> Vec<Box<dyn Hittable>> {
    let mut world:Vec<Box<dyn Hittable>> = Vec::new();

    let material_ground = Material::Lambertian(Lambertian::new(Color::new(0.5, 0.5, 0.5)));
    let ground = Sphere {
        center: Point3::new(0.0, -1000.0, 0.0),
        radius: 1000.0,
        material: material_ground
    };
    world.push(Box::new(ground));

    let unit_range = Uniform::from(0.0..1.0);
    let fuzz_range = Uniform::from(0.0..0.5);
    let mut rng = rand::thread_rng();
    let p = Point3::new(4.0, 0.2, 0.0);
    for a in -1..11 {
        for b in -11..11 {
            let choose_material = unit_range.sample(&mut rng);
            let center = Point3::new(
                (a as f64) + 0.9*unit_range.sample(&mut rng),
                0.2,
                (b as f64) + 0.9*unit_range.sample(&mut rng));
            if (center - p).length() < 0.9 {
                continue;
            }
            let material = match choose_material {
                _ if choose_material < 0.8 => Material::Lambertian(Lambertian::new(
                    Color::random(0.0, 1.0) * Color::random(0.0, 1.0))),
                _ if choose_material < 0.95 => Material::Metal(Metal::new(
                    Color::random(0.5, 1.0),
                    fuzz_range.sample(&mut rng))),
                _ => Material::Dielectric(Dielectric::new(1.5)),
            };
            let sphere = Box::new(Sphere {
                center,
                radius: 0.2,
                material
            });
            world.push(sphere);
        }
    }

    let material1 = Material::Dielectric(Dielectric::new(1.5));
    world.push(Box::new(Sphere {
        center: Point3::new(0.0, 1.0, 0.0),
        radius: 1.0,
        material: material1
    }));
    let material2 = Material::Lambertian(Lambertian::new(Color::new(0.4, 0.2, 0.1)));
    world.push(Box::new(Sphere {
        center: Point3::new(-4.0, 1.0, 0.0),
        radius: 1.0,
        material: material2
    }));
    let material3 = Material::Metal(Metal::new(Color::new(0.7, 0.6, 0.5), 0.0));
    world.push(Box::new(Sphere {
        center: Point3::new(4.0, 1.0, 0.0),
        radius: 1.0,
        material: material3
    }));

    world
}

fn main() {
    // Image
    const ASPECT_RATIO: f64 = 3.0 / 2.0;
    const IMAGE_WIDTH: usize = 1200;
    const IMAGE_HEIGHT: usize = (IMAGE_WIDTH as f64 / ASPECT_RATIO) as usize;
    const SAMPLES_PER_PIXEL: i32 = 100;
    const MAX_DEPTH: i32 = 50;

    let look_from = Point3::new(13.0, 2.0, 3.0);
    let look_at = Point3::new(0.0, 0.0, 0.0);
    let focus_dist = 10.0;
    // Camera
    let cam = Camera::new(
        look_from,
        look_at,
        Vec3::new(0.0, 1.0, 0.0),
        ASPECT_RATIO,
        20.0,
        0.1,
        focus_dist);

    // World
    let world= random_scene();

    //println!("P3\n{} {}\n255", IMAGE_WIDTH, IMAGE_HEIGHT);

    let between = Uniform::from(0.0..1.0);
    let mut pixels = Vec::<u8>::new();
    let mut rng = rand::thread_rng();
    for j in (0..IMAGE_HEIGHT).rev() {
        eprint!("\rScanlines remaining: {} ", j);
        for i in 0..IMAGE_WIDTH {
            let mut color = Color::default();
            (0..SAMPLES_PER_PIXEL).for_each(|_s| {
                let random_number = between.sample(&mut rng);
                let u = (i as f64 + random_number) / (IMAGE_WIDTH - 1) as f64;
                let v = (j as f64 + random_number) / (IMAGE_HEIGHT - 1) as f64;
                let ray = cam.get_ray(u, v);
                color += ray.pixel_color(&world, MAX_DEPTH);
            });
            pixels.append(&mut color.into_bytes(SAMPLES_PER_PIXEL));

            //color.write_color(SAMPLES_PER_PIXEL);
        }
    }
    P3::write("i.ppm", &pixels, IMAGE_WIDTH, IMAGE_HEIGHT).expect("Error writing image: {}");
    eprintln!("\nDone");
}
